use proc_macro::TokenStream;
use syn::{parse_macro_input, DeriveInput};
use quote::{quote,format_ident};

use convert_case::{Case, Casing};

fn parse_microrm_ref(attrs: &Vec<syn::Attribute>) -> proc_macro2::TokenStream {
    for attr in attrs {
        if attr.path.segments.len() == 0 { continue }

        if attr.tokens.is_empty() {
            if attr.path.segments.last().unwrap().ident == "microrm_internal" {
                return quote!{ crate }.into()
            }
        }
    }
    
    quote!{ ::microrm }
}

/// Turns a serializable/deserializable struct into a microrm entity model.
///
/// There are two important visible effects:
/// - Provides an implementation of `microrm::model::Entity`
/// - Defines a <struct-name>Columns enum
///
/// Note that names are converted from CamelCase to snake_case and vice versa
/// where applicable, so a struct named `TestModel` is given a table name `test_model`
/// and a struct field named `field_name` is given a variant name of `FieldName`.
///
/// The `#[microrm...]` attributes can be used to control the derivation somewhat.
/// The following are understood for the Entity struct:
/// - `#[microrm_internal]`: this is internal to the microrm crate (of extremely limited usefulness
/// outside the microrm library)
/// The following are understood on individual fields
/// - `#[microrm_foreign]`: this is a foreign key (and must be an type implementing `EntityID`)
#[proc_macro_derive(Entity, attributes(microrm_internal,microrm_foreign))]
pub fn derive_entity(tokens: TokenStream) -> TokenStream {
    let input = parse_macro_input!(tokens as DeriveInput);

    let mut microrm_ref = parse_microrm_ref(&input.attrs);

    let struct_name = &input.ident;
    let enum_name = format_ident!("{}Columns", &input.ident);
    let id_name = format_ident!("{}ID", &input.ident);

    let table_name = format!("{}", struct_name).to_case(Case::Snake);

    let st = match input.data {
        syn::Data::Struct(st) => st,
        _ => panic!("Can only use derive(Entity) on structs!")
    };
    let fields = match st.fields {
        syn::Fields::Named(fields) => fields,
        _ => panic!("Can only use derive(Entity) on non-unit structs with named fields!")
    };

    for name in fields.named.iter() {
        // println!("ty: {:?}", name.ty);
    }

    let mut variants = syn::punctuated::Punctuated::<syn::Ident, syn::token::Comma>::new();
    let mut field_names = syn::punctuated::Punctuated::<proc_macro2::TokenStream, syn::token::Comma>::new();
    let mut value_references = syn::punctuated::Punctuated::<proc_macro2::TokenStream, syn::token::Comma>::new();
    for name in fields.named.iter() {
        let converted_case = format!("{}", name.ident.as_ref().unwrap().clone()).to_case(Case::UpperCamel);
        let converted_case = format_ident!("{}", converted_case);
        variants.push(converted_case.clone());

        let field_name = name.ident.as_ref().unwrap().clone();
        let field_name_str = format!("{}", field_name);
        field_names.push(quote!{ Self::Column::#converted_case => #field_name_str }.into());

        value_references.push(quote!{ &self. #field_name });
    }

    let field_count = fields.named.iter().count();

    let ret = quote!{
        // Related types for #struct_name
        #[derive(Clone,Copy)]
        #[allow(unused)]
        pub enum #enum_name {
            #variants
        }

        #[derive(Debug,PartialEq,Clone,Copy,#microrm_ref::re_export::serde::Serialize,#microrm_ref::re_export::serde::Deserialize)]
        #[allow(unused)]
        pub struct #id_name (i64);

        // Implementations for related types
        impl #microrm_ref::model::EntityColumns for #enum_name {
            type Entity = #struct_name;
        }

        impl #microrm_ref::model::EntityID for #id_name {
            type Entity = #struct_name;
            fn from_raw_id(raw: i64) -> Self { Self(raw) }
            fn raw_id(&self) -> i64 { self.0 }
        }

        impl #microrm_ref::re_export::rusqlite::ToSql for #id_name {
            fn to_sql(&self) -> Result<#microrm_ref::re_export::rusqlite::types::ToSqlOutput<'_>, #microrm_ref::re_export::rusqlite::Error> {
                self.0.to_sql()
            }
        }

        // Implementations for #struct_name
        impl #microrm_ref::model::Entity for #struct_name {
            type Column = #enum_name;
            type ID = #id_name;

            fn table_name() -> &'static str { #table_name }
            fn column_count() -> usize {
                #field_count
            }
            fn index(c: Self::Column) -> usize {
                c as usize
            }
            fn name(c: Self::Column) -> &'static str {
                match c {
                    #field_names
                }
            }
            fn values(&self) -> Vec<&dyn #microrm_ref::re_export::rusqlite::ToSql> {
                vec![ #value_references ]
            }
        }
    }.into();

    ret
}

#[proc_macro_derive(Modelable, attributes(microrm_internal))]
pub fn derive_modelable(tokens: TokenStream) -> TokenStream {
    let input = parse_macro_input!(tokens as DeriveInput);

    let mut microrm_ref = parse_microrm_ref(&input.attrs);

    let ident = input.ident;

    quote!{
        impl #microrm_ref::re_export::rusqlite::ToSql for #ident {
            fn to_sql(&self) -> #microrm_ref::re_export::rusqlite::Result<#microrm_ref::re_export::rusqlite::types::ToSqlOutput<'_>> {
                use #microrm_ref::re_export::rusqlite::types::{ToSqlOutput,Value};
                Ok(ToSqlOutput::Owned(Value::Text(#microrm_ref::re_export::serde_json::to_string(self).expect("can be serialized"))))
            }
        }
    }.into()
}
